/**
    This is a synthetic generator to create problem instances for various
    generalized network design problems.

    Copyright (C) 2013  Author: Yexiang Xue

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. 

    Contact Yexiang Xue at yexiang@cs.cornell.edu
            Bistra Dilkina at bistra@cs.cornell.edu
 */
#include <algorithm>
#include <cmath>
//#include <random>

#include <cstdio>
#include <ctime>
#include <vector>
#include <cstring>

using namespace std;

const int NODATA = -9999;

const int ResistanceRange = 2000;
const int CostRange = 2000;

const int randomNoiseLvl = 10;
const double extraPeakContrib = 0.5;
const double termPeakContrib = 1;

const int STRUCTURED = 1;
const int RANDOM = 0;

int method = STRUCTURED;  // 1 for structured; 0 for randomized.

double sigma_center;
int num_extra;

char* inputfilename = "gaussian_multi_core_input.txt";

char filename[100];
int n;
int sp; 
int blob_size;
int num_species;
int num_core;   // how many core pairs per species;
int* x_axis;
int* y_axis;

int numCopy; // how many random instances we need;

typedef int** Matrix;

void read_input() {
  // TODO: incorporate method 
  printf("%s\n", inputfilename);
  FILE* inp = fopen(inputfilename, "r");

  fscanf(inp, "%s", filename);
  
  char methodstring[100];
  fscanf(inp, "%s", methodstring);
  if (strcmp(methodstring, "STRUCTURED") == 0)
    method = STRUCTURED;
  else
    method = RANDOM;

  fscanf(inp, "%d %d %d %d %d", &n, &sp, &blob_size, &num_species,
                                &num_core);
  
  /*int i;
  x_axis = new int[num_species*num_core];
  y_axis = new int[num_species*num_core];
  for (i = 0; i < num_species*num_core; ++ i) {
    fscanf(inp, "%d %d", x_axis + i, y_axis + i);
    //printf("%d %d\n", x_axis[i], y_axis[i]);
  }*/

  fscanf(inp, "%d", &numCopy);
  //printf("%d\n", numCopy);
  
  if (method == STRUCTURED){
    fscanf(inp, "%d", &num_extra);
    fscanf(inp, "%lf", &sigma_center);
    //printf("num_extra=%d sigma_center=%lf\n", num_extra, sigma_center);
  }


  fclose(inp);
}


Matrix genMatrix(int m, int n/*, int initial*/) {
  Matrix a = new int*[m];
  int i,j;
  for (i = 0; i < m; ++ i){
    a[i] = new int[n];
    //memset(a[i], initial, sizeof(a[i]));
    //for (j = 0; j < n; ++ j)
    //  a[i][j] = initial;
  }

  return a;
}

void outputMatrix(Matrix a, const char* filename) {
    FILE* fid = fopen(filename, "w");

    fprintf(fid, "ncols         %d\n", n);
    fprintf(fid, "nrows         %d\n", n);
    fprintf(fid, "xllcorner     111865.9531\n");
    fprintf(fid, "yllcorner     17726.070300002\n");
    fprintf(fid, "cellsize      6000\n");
    fprintf(fid, "NODATA_value  -9999\n");

    int i,j;
    for (i = 0; i < n; ++ i){
      for (j = 0; j < n; ++ j)
        fprintf(fid, "%d ", a[i][j]);
      fprintf(fid, "\n");
    }

    fclose(fid);
}


const int step[4][2] = {{1,0}, {0,1}, {-1,0}, {0,-1}};

void createBlob(Matrix term, int px, int py, int blobsize, int fill) {
  int len = (n/sp);
  
  int rlow = len*px;
  int rhigh = rlow + len;  // [rlow, rhigh)
  int clow = len*py;
  int chigh = clow + len;

  int *x = new int[blobsize];
  int *y = new int[blobsize];

  //x[0] = rlow + rand() % len;
  //y[0] = clow + rand() % len;
  x[0] = rlow + len / 2;
  y[0] = clow + len / 2;
  term[x[0]][y[0]] = fill;

  int tp1 = -1, tp2 = 0;
  int num = 1, i;
  int xx, yy;
  while (tp1 < tp2 && num < blobsize) {
    ++tp1;
    for (i = 0; i < 4; ++ i){
      xx = x[tp1] + step[i][0];
      yy = y[tp1] + step[i][1];
      if (xx >= rlow && xx < rhigh &&
          yy >= clow && yy < chigh && 
          term[xx][yy] == NODATA) {
        term[xx][yy] = fill;
        ++tp2;
        x[tp2] = xx;
        y[tp2] = yy;
        ++num;
      }
      if (num == blobsize)
        break;
    }
  }

  delete [] x;
  delete [] y;
}


void genTermPair(Matrix term, int which) {
  int xx;
  int yy;
  int x;
  int y;

  int i;
  for (i = 0; i < num_core; ++ i) {
    x = x_axis[which*num_core+i];
    y = y_axis[which*num_core+i];
    xx = sp - x - 1;
    yy = sp - y - 1;
    createBlob(term, x, y, blob_size, i + i);
    createBlob(term, xx, yy, blob_size, i + i + 1);
  }
}

void outputTermPairFile(const char* filename) {
  char filename1[200];
  // change extension name to txt
  strcpy(filename1, filename);
  int len = strlen(filename1);
  filename1[len-1] = 't';
  filename1[len-2] = 'x';
  filename1[len-3] = 't';
  FILE* oup = fopen(filename1, "w");

  fprintf(oup, "%d\n", num_core);
  for (int i = 0; i < num_core; ++ i) {
    fprintf(oup, "%d %d\n", i+i, i+i+1);
  }  
  fclose(oup);
}


inline double normalpdf(int x, int y, int mu_x, int mu_y, 
                 double sigmax, double sigmay, double rho) {
  return exp(-((x-mu_x)*(x-mu_x)/sigmax/sigmax + 
               (y-mu_y)*(y-mu_y)/sigmay/sigmay -
               2*rho*(x-mu_x)*(y-mu_y)/sigmax/sigmay)/2/(1-rho*rho))
    /2.0/3.1415927/sigmax/sigmay/sqrt(1-rho*rho);
}

inline double rand_float(double x, double y) {
  return x + double(rand()) / double(RAND_MAX) * (y-x);
}

/** get gaussian-like resistance
 */
void genResistance(Matrix r, int which) {
  int i,j;
  int k;

  //printf("num_extra= %d\n", num_extra);

  // generate randomly peaked blobs
  int* x_extra = new int[num_extra];
  int* y_extra = new int[num_extra];
  int low_n = n / 5;
  int high_n = n - low_n;
  int range_n = high_n - low_n;
  for (k = 0; k < num_extra; ++ k){
    x_extra[k] = low_n + rand() % range_n;
    y_extra[k] = low_n + rand() % range_n;
  }
  
  double sigma_x = sigma_center * rand_float(1.0, 1.5);
  double sigma_y = sigma_center * rand_float(1.0, 1.5);
  
  double* rho = new double[num_extra];
  double* extra_normalize = new double[num_extra];

  for (k = 0; k < num_extra; ++ k){
    rho[k] = rand_float(-0.8, 0.8);
    //printf("rho[%d]= %lf\n", k, rho[k]);
    extra_normalize[k] = normalpdf(0, 0, 0, 0, sigma_x, sigma_y, rho[k]);
    //printf("extra_normalize[%d]= %lf\n", k, extra_normalize[k]);
  }
  
  double blob_normalize = normalpdf(0, 0, 0, 0, sigma_center, sigma_center,
                                    0);
  //printf("blob_normalize= %lf\n", blob_normalize);
  // 
  for (i = 0; i < n; ++ i)
    for (j = 0;j < n; ++ j){
      r[i][j] = rand() % (ResistanceRange / randomNoiseLvl);
      for (k = 0; k < num_core; ++ k) {
        /*printf("pdf= %lf\n",normalpdf(i,j,
                                 x_axis[which*num_core+i]*(n/sp)+(n/sp/2),
                                 y_axis[which*num_core+i]*(n/sp)+(n/sp/2),
                                 sigma_center, sigma_center, 0));*/
        int xloc = x_axis[which*num_core+k]*(n/sp)+(n/sp/2);
        int yloc = y_axis[which*num_core+k]*(n/sp)+(n/sp/2);
        r[i][j] += int(ResistanceRange * termPeakContrib * 
                       normalpdf(i,j,
                                 xloc,
                                 yloc,
                                 sigma_center, sigma_center, 0)
                       / blob_normalize);
        xloc = n - xloc;
        yloc = n - yloc;
        r[i][j] += int(ResistanceRange * termPeakContrib * 
                       normalpdf(i,j,
                                 xloc,
                                 yloc,
                                 sigma_center, sigma_center, 0)
                       / blob_normalize);
        
      }
      for (k = 0; k < num_extra; ++ k) {
        r[i][j] += int(ResistanceRange * extraPeakContrib * 
                       normalpdf(i,j,
                                 x_extra[k],y_extra[k],
                                 sigma_x, sigma_y, rho[k])
                       / extra_normalize[k]);
      }

      if (r[i][j] >= ResistanceRange)
        r[i][j] = ResistanceRange-1;
      if (r[i][j] < 0)
        r[i][j] = 0;

	 r[i][j] = ResistanceRange - r[i][j];
    }

  delete [] x_extra;
  delete [] y_extra;
  delete [] rho;
  delete [] extra_normalize;
}


void combineMaskTermPair(Matrix mask, Matrix term) {
  int i,j;
  for (i = 0; i < n; ++ i)
    for (j = 0; j < n; ++ j)
      if (term[i][j] != NODATA)
        mask[i][j] = 0;
}


char* combineName(char* filename, const char* ext, int no = NODATA) {
  char* ans = new char[100];
  //"_";
  strcpy(ans, filename);
  if (strcmp(ext, "") != 0){
    strcat(ans, "_");
    strcat(ans, ext);
  }
  if (no != NODATA){
    strcat(ans, "_");
    char num_no[50];
    sprintf(num_no, "%d", no);
    strcat(ans, num_no);
  }
  if (strcmp(ext, "") != 0) {
    strcat(ans, ".asc");
  }
  return ans;
}

void genResistance(Matrix r) {
  int i,j;
  for (i = 0; i < n; ++ i)
    for (j = 0;j < n; ++ j)
      r[i][j] = rand() % ResistanceRange;
}

void genCost(Matrix cost, Matrix mask) {
  int i,j;
  for (i = 0; i < n; ++ i)
    for (j = 0;j < n; ++ j)
      if (mask[i][j] == 0)
        cost[i][j] = 0;
      else
        cost[i][j] = rand() % CostRange;  
}

void genCost(Matrix cost, Matrix mask, Matrix* res, int num_species,
             Matrix tmp) {
  int i,j,k;
  int big = -100000000;
  for (i = 0; i < n; ++ i)
    for (j = 0; j < n; ++ j){
      tmp[i][j] = 2147483647;
      for (k = 0; k < num_species; ++ k){
        if (res[k][i][j] < tmp[i][j])
          tmp[i][j] = res[k][i][j];
      }
      tmp[i][j] += rand() % (CostRange / randomNoiseLvl);
      if (tmp[i][j] > big)
        big = tmp[i][j];
    }
  for (i = 0; i < n; ++ i)
    for (j = 0; j < n; ++ j)
      tmp[i][j] = tmp[i][j] / double(big) * CostRange;
      
  for (i = 0; i < n; ++ i)
    for (j = 0;j < n; ++ j)
      if (mask[i][j] == 0)
        cost[i][j] = 0;
      else{
        cost[i][j] = tmp[i][j]; 
      }
}


struct sort_pair {
  int a, b;
};

bool sort_pair_less_wrt_first(const sort_pair& x, 
                              const sort_pair& y) {
  return x.a < y.a;
}

void getRank(int* rank, int n) {
  int i;
  sort_pair* k = new sort_pair[n];
  for (i = 0; i < n; ++ i) {
    k[i].a = rand();
    k[i].b = i;
  }
  sort(k, k+n, sort_pair_less_wrt_first);
  for (i = 0; i < n; ++ i)
    rank[i] = k[i].b;
  delete [] k;
}

void cleanMatrix(Matrix a, int m, int n, int value) {
  int i,j;
  for (i = 0; i < m; ++ i)
    for (j = 0; j < n; ++ j)
      a[i][j] = value;
}


bool hasSame(int* x, int* y, int n) {
  int i;
  for (i = 0; i < n; ++ i)
    if (x[i] == x[n] && y[i] == y[n])
      return true;
  return false;
}

void genCoreLoc(int* xloc, int* yloc, int n, int split) {
  printf("Start genCoreLoc\n");
  int range = split / 2;
  int i;
  int randtmp;

  for (i = 0; i < n; ++ i){
    do {
      randtmp = rand() % 4;
      if (randtmp == 0){
        // left edge
        yloc[i] = 0;
        xloc[i] = rand() % range;
      }else if (randtmp == 4){
        // right edge
        xloc[i] = rand() % range;
        yloc[i] = split - 1;
      }else{
        // above edge
        xloc[i] = 0;
        yloc[i] = rand() % split;
      }
    }while (hasSame(xloc, yloc, i));
    printf("%d %d\n", xloc[i], yloc[i]);
  }
}


int main(int argc, char** argv) {

  if (argc > 1)
    inputfilename = argv[1];

  srand(time(NULL));
  read_input();

  printf("num_species= %d\n", num_species);

  int i,j;
  int* rank = new int[num_species];
  
  Matrix *termpair = new Matrix[num_species];
  Matrix *res = new Matrix[num_species];
  Matrix mask = genMatrix(n, n);
  Matrix tmpMatrix = genMatrix(n, n);
  Matrix cost = genMatrix(n, n);
  for (i = 0; i < num_species; ++ i){
    termpair[i] = genMatrix(n, n);
    res[i] = genMatrix(n, n);
  }

  x_axis = new int[num_species*num_core];
  y_axis = new int[num_species*num_core];

  //printf("numCopy=%d\n", numCopy);
  for (j = 0; j < numCopy; ++ j) {
    // get core area location
    genCoreLoc(x_axis, y_axis, num_species*num_core, sp);

    getRank(rank, num_species);
    
    char* filename1 = combineName(filename, "", j);
    cleanMatrix(mask, n, n, NODATA);

    for (i = 0; i < num_species; ++ i) {
      cleanMatrix(termpair[i], n, n, NODATA);
      genTermPair(termpair[i], rank[i]);
      combineMaskTermPair(mask, termpair[i]);

      outputMatrix(termpair[i], combineName(filename1, "term", i));
      outputTermPairFile(combineName(filename1, "termpair", i));
    }

  
    for (i = 0; i < num_species; ++ i) {
      cleanMatrix(res[i], n, n, 0);
      if (method == STRUCTURED)
        genResistance(res[i], rank[i]);
      else
        genResistance(res[i]);
      outputMatrix(res[i], combineName(filename1, "res", i));
    }

  
    cleanMatrix(cost, n, n, 0);
    if (method == STRUCTURED)
      genCost(cost, mask, res, num_species, tmpMatrix);
    else
      genCost(cost, mask);
    outputMatrix(cost, combineName(filename1, "cost"));
    
    delete [] filename1;
  }

  delete [] rank;
  return 0;
}

